
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "emo.h"

int main(int argc, char **argv) {
  int i, j, size, dim, *filter, type;
  double *data, *z, *t, *alpha;
  EMO_Rand rnd;

  if(argc != 4) {
    printf("Syntax: %s size dimension dominance_type[1:weak, 2:strict, 3:strong, 4:alpha]\n", argv[0]);
    exit(1);
  }

  size = atoi(argv[1]);
  dim = atoi(argv[2]);
  type = atoi(argv[3]);

  EMO_Rand_alloc_from_file(&rnd, NULL, "./input/seed.dat", 0);

  filter = (int *) malloc(sizeof(int) * size);
  data = (double *) malloc(sizeof(double) * size * dim);
  z = (double *) malloc(sizeof(double) * dim);

  /* Initialize data */
  for(i = 0; i < size; i++)
    for(j = 0; j < dim; j++)
      data[i * dim + j] = EMO_Rand_prob1(&rnd);

  /* Reference point */
  for(j = 0; j < dim; j++)
    z[j] = 0.5;

  EMO_File_write(data, NULL, size, dim, "all.txt", "%.3f ");
  memset(filter, 0, sizeof(int) * size);


  switch(type) {
    case 1:
              for(i = 0; i < size; i++)
                if(EMO_Dominance_weak(z, data + i*dim, dim) == 1)
                  filter[i] = 1; 
              break;
    case 2:
              for(i = 0; i < size; i++)
                if(EMO_Dominance_strict(z, data + i*dim, dim) == 1)
                  filter[i] = 1; 
              break;
    case 3:
              for(i = 0; i < size; i++)
                if(EMO_Dominance_strong(z, data + i*dim, dim) == 1)
                  filter[i] = 1; 
              break;
     case 4:

              alpha = (double *) malloc(sizeof(double) * dim * dim);
              t = (double *) malloc(sizeof(double) * dim);

              for(j = dim*dim - 1; j > -1; j--)
                alpha[j] = 0.1;

              for(i = 0; i < size; i++)
                if(EMO_Dominance_alpha(z, data + i*dim, t, alpha, dim) == 1)
                  filter[i] = 1; 

              free(alpha);
              free(t);
              break;

     default: printf("Error, unknown dominance.\n");
              return 1;

  }

  EMO_File_write(data, filter, size, dim, "dominated.txt", "%.3f ");

  EMO_Rand_free(&rnd);

  free(filter);
  free(data);
  free(z);

  return 0;
}

